/*
    FreeRTOS V7.6.0 - Copyright (C) 2013 Real Time Engineers Ltd. 
    All rights reserved

    VISIT http://www.FreeRTOS.org TO ENSURE YOU ARE USING THE LATEST VERSION.

    ***************************************************************************
     *                                                                       *
     *    FreeRTOS provides completely free yet professionally developed,    *
     *    robust, strictly quality controlled, supported, and cross          *
     *    platform software that has become a de facto standard.             *
     *                                                                       *
     *    Help yourself get started quickly and support the FreeRTOS         *
     *    project by purchasing a FreeRTOS tutorial book, reference          *
     *    manual, or both from: http://www.FreeRTOS.org/Documentation        *
     *                                                                       *
     *    Thank you!                                                         *
     *                                                                       *
    ***************************************************************************

    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation >>!AND MODIFIED BY!<< the FreeRTOS exception.

    >>! NOTE: The modification to the GPL is included to allow you to distribute
    >>! a combined work that includes FreeRTOS without being obliged to provide
    >>! the source code for proprietary components outside of the FreeRTOS
    >>! kernel.

    FreeRTOS is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  Full license text is available from the following
    link: http://www.freertos.org/a00114.html

    1 tab == 4 spaces!

    ***************************************************************************
     *                                                                       *
     *    Having a problem?  Start by reading the FAQ "My application does   *
     *    not run, what could be wrong?"                                     *
     *                                                                       *
     *    http://www.FreeRTOS.org/FAQHelp.html                               *
     *                                                                       *
    ***************************************************************************

    http://www.FreeRTOS.org - Documentation, books, training, latest versions,
    license and Real Time Engineers Ltd. contact details.

    http://www.FreeRTOS.org/plus - A selection of FreeRTOS ecosystem products,
    including FreeRTOS+Trace - an indispensable productivity tool, a DOS
    compatible FAT file system, and our tiny thread aware UDP/IP stack.

    http://www.OpenRTOS.com - Real Time Engineers ltd license FreeRTOS to High
    Integrity Systems to sell under the OpenRTOS brand.  Low cost OpenRTOS
    licenses offer ticketed support, indemnification and middleware.

    http://www.SafeRTOS.com - High Integrity Systems also provide a safety
    engineered and independently SIL3 certified version for use in safety and
    mission critical applications that require provable dependability.

    1 tab == 4 spaces!
*/

/* 

Changes from V2.6.0

	+ AVR port - Replaced the inb() and outb() functions with direct memory
	  access.  This allows the port to be built with the 20050414 build of
	  WinAVR.
*/


#include <stdlib.h>
#include <stdbool.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/boot.h>

#include "FreeRTOS.h"
#include "task.h"

#if configUSE_TICKLESS_IDLE_DEBUG > 0
#	include <stdio.h>
#	include <avr/pgmspace.h>
#endif

/*-----------------------------------------------------------
 * Implementation of functions defined in portable.h for the AVR port.
 *----------------------------------------------------------*/

/* Start tasks with interrupts enables. */
#define portFLAGS_INT_ENABLED					( ( portSTACK_TYPE ) 0x80 )

/* Hardware constants for timer 1. */
#define portCLEAR_COUNTER_ON_MATCH				( ( unsigned char ) 0x08 )
#define portPRESCALE_64							( ( unsigned char ) 0x03 )
#define portCLOCK_PRESCALER						( ( unsigned long ) 64 )
#define portCOMPARE_MATCH_A_INTERRUPT_ENABLE	( ( unsigned char ) 0x10 )

/*-----------------------------------------------------------*/

/* We require the address of the pxCurrentTCB variable, but don't want to know
any details of its type. */
typedef void tskTCB;
extern volatile tskTCB * volatile pxCurrentTCB;

/*-----------------------------------------------------------*/

/* 
 * Macro to save all the general purpose registers, the save the stack pointer
 * into the TCB.  
 * 
 * The first thing we do is save the flags then disable interrupts.  This is to 
 * guard our stack against having a context switch interrupt after we have already 
 * pushed the registers onto the stack - causing the 32 registers to be on the 
 * stack twice. 
 * 
 * r1 is set to zero as the compiler expects it to be thus, however some
 * of the math routines make use of R1. 
 * 
 * The interrupts will have been disabled during the call to portSAVE_CONTEXT()
 * so we need not worry about reading/writing to the stack pointer. 
 */

#define portSAVE_CONTEXT()									\
	asm volatile (	"push	r0						\n\t"	\
					"in		r0, __SREG__			\n\t"	\
					"cli							\n\t"	\
					"push	r0						\n\t"	\
					"push	r1						\n\t"	\
					"clr	r1						\n\t"	\
					"push	r2						\n\t"	\
					"push	r3						\n\t"	\
					"push	r4						\n\t"	\
					"push	r5						\n\t"	\
					"push	r6						\n\t"	\
					"push	r7						\n\t"	\
					"push	r8						\n\t"	\
					"push	r9						\n\t"	\
					"push	r10						\n\t"	\
					"push	r11						\n\t"	\
					"push	r12						\n\t"	\
					"push	r13						\n\t"	\
					"push	r14						\n\t"	\
					"push	r15						\n\t"	\
					"push	r16						\n\t"	\
					"push	r17						\n\t"	\
					"push	r18						\n\t"	\
					"push	r19						\n\t"	\
					"push	r20						\n\t"	\
					"push	r21						\n\t"	\
					"push	r22						\n\t"	\
					"push	r23						\n\t"	\
					"push	r24						\n\t"	\
					"push	r25						\n\t"	\
					"push	r26						\n\t"	\
					"push	r27						\n\t"	\
					"push	r28						\n\t"	\
					"push	r29						\n\t"	\
					"push	r30						\n\t"	\
					"push	r31						\n\t"	\
					"lds	r26, pxCurrentTCB		\n\t"	\
					"lds	r27, pxCurrentTCB + 1	\n\t"	\
					"in		r0, 0x3d				\n\t"	\
					"st		x+, r0					\n\t"	\
					"in		r0, 0x3e				\n\t"	\
					"st		x+, r0					\n\t"	\
				);

/* 
 * Opposite to portSAVE_CONTEXT().  Interrupts will have been disabled during
 * the context save so we can write to the stack pointer. 
 */

#define portRESTORE_CONTEXT()								\
	asm volatile (	"lds	r26, pxCurrentTCB		\n\t"	\
					"lds	r27, pxCurrentTCB + 1	\n\t"	\
					"ld		r28, x+					\n\t"	\
					"out	__SP_L__, r28			\n\t"	\
					"ld		r29, x+					\n\t"	\
					"out	__SP_H__, r29			\n\t"	\
					"pop	r31						\n\t"	\
					"pop	r30						\n\t"	\
					"pop	r29						\n\t"	\
					"pop	r28						\n\t"	\
					"pop	r27						\n\t"	\
					"pop	r26						\n\t"	\
					"pop	r25						\n\t"	\
					"pop	r24						\n\t"	\
					"pop	r23						\n\t"	\
					"pop	r22						\n\t"	\
					"pop	r21						\n\t"	\
					"pop	r20						\n\t"	\
					"pop	r19						\n\t"	\
					"pop	r18						\n\t"	\
					"pop	r17						\n\t"	\
					"pop	r16						\n\t"	\
					"pop	r15						\n\t"	\
					"pop	r14						\n\t"	\
					"pop	r13						\n\t"	\
					"pop	r12						\n\t"	\
					"pop	r11						\n\t"	\
					"pop	r10						\n\t"	\
					"pop	r9						\n\t"	\
					"pop	r8						\n\t"	\
					"pop	r7						\n\t"	\
					"pop	r6						\n\t"	\
					"pop	r5						\n\t"	\
					"pop	r4						\n\t"	\
					"pop	r3						\n\t"	\
					"pop	r2						\n\t"	\
					"pop	r1						\n\t"	\
					"pop	r0						\n\t"	\
					"out	__SREG__, r0			\n\t"	\
					"pop	r0						\n\t"	\
				);

/*-----------------------------------------------------------*/

/*
 * Perform hardware setup to enable ticks from timer 1, compare match A.
 */
static void prvSetupTimerInterrupt( void );
/*-----------------------------------------------------------*/

/*
 * Setup timer 0 async mode with low-freq crystal.
 */
#if configUSE_TICKLESS_IDLE == 1
	static void prvSetupTimerAsync( void );
#endif

/* 
 * See header file for description. 
 */
portSTACK_TYPE *pxPortInitialiseStack( portSTACK_TYPE *pxTopOfStack, pdTASK_CODE pxCode, void *pvParameters )
{
unsigned short usAddress;

	/* Place a few bytes of known values on the bottom of the stack. 
	This is just useful for debugging. */

	*pxTopOfStack = 0x11;
	pxTopOfStack--;
	*pxTopOfStack = 0x22;
	pxTopOfStack--;
	*pxTopOfStack = 0x33;
	pxTopOfStack--;

	/* Simulate how the stack would look after a call to vPortYield() generated by 
	the compiler. */

	/*lint -e950 -e611 -e923 Lint doesn't like this much - but nothing I can do about it. */

	/* The start of the task code will be popped off the stack last, so place
	it on first. */
	usAddress = ( unsigned short ) pxCode;
	*pxTopOfStack = ( portSTACK_TYPE ) ( usAddress & ( unsigned short ) 0x00ff );
	pxTopOfStack--;

	usAddress >>= 8;
	*pxTopOfStack = ( portSTACK_TYPE ) ( usAddress & ( unsigned short ) 0x00ff );
	pxTopOfStack--;

	/* Next simulate the stack as if after a call to portSAVE_CONTEXT().  
	portSAVE_CONTEXT places the flags on the stack immediately after r0
	to ensure the interrupts get disabled as soon as possible, and so ensuring
	the stack use is minimal should a context switch interrupt occur. */
	*pxTopOfStack = ( portSTACK_TYPE ) 0x00;	/* R0 */
	pxTopOfStack--;
	*pxTopOfStack = portFLAGS_INT_ENABLED;
	pxTopOfStack--;


	/* Now the remaining registers.   The compiler expects R1 to be 0. */
	*pxTopOfStack = ( portSTACK_TYPE ) 0x00;	/* R1 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x02;	/* R2 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x03;	/* R3 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x04;	/* R4 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x05;	/* R5 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x06;	/* R6 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x07;	/* R7 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x08;	/* R8 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x09;	/* R9 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x10;	/* R10 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x11;	/* R11 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x12;	/* R12 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x13;	/* R13 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x14;	/* R14 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x15;	/* R15 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x16;	/* R16 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x17;	/* R17 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x18;	/* R18 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x19;	/* R19 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x20;	/* R20 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x21;	/* R21 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x22;	/* R22 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x23;	/* R23 */
	pxTopOfStack--;

	/* Place the parameter on the stack in the expected location. */
	usAddress = ( unsigned short ) pvParameters;
	*pxTopOfStack = ( portSTACK_TYPE ) ( usAddress & ( unsigned short ) 0x00ff );
	pxTopOfStack--;

	usAddress >>= 8;
	*pxTopOfStack = ( portSTACK_TYPE ) ( usAddress & ( unsigned short ) 0x00ff );
	pxTopOfStack--;

	*pxTopOfStack = ( portSTACK_TYPE ) 0x26;	/* R26 X */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x27;	/* R27 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x28;	/* R28 Y */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x29;	/* R29 */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x30;	/* R30 Z */
	pxTopOfStack--;
	*pxTopOfStack = ( portSTACK_TYPE ) 0x031;	/* R31 */
	pxTopOfStack--;

	/*lint +e950 +e611 +e923 */

	return pxTopOfStack;
}
/*-----------------------------------------------------------*/



portBASE_TYPE xPortStartScheduler( void )
{
	/* Setup the hardware to generate the tick. */
	prvSetupTimerInterrupt();

	#if configUSE_TICKLESS_IDLE == 1
		prvSetupTimerAsync();
	#endif

	/* Restore the context of the first task that is going to run. */
	portRESTORE_CONTEXT();

	/* Simulate a function call end as generated by the compiler.  We will now
	jump to the start of the task the context of which we have just restored. */
	asm volatile ( "ret" );

	/* Should not get here. */
	return pdTRUE;
}
/*-----------------------------------------------------------*/

void vPortEndScheduler( void )
{
	/* It is unlikely that the AVR port will get stopped.  If required simply
	disable the tick interrupt here. */
}
/*-----------------------------------------------------------*/

/*
 * Manual context switch.  The first thing we do is save the registers so we
 * can use a naked attribute.
 */
void vPortYield( void ) __attribute__ ( ( naked ) );
void vPortYield( void )
{
	portSAVE_CONTEXT();
	vTaskSwitchContext();
	portRESTORE_CONTEXT();

	asm volatile ( "ret" );
}
/*-----------------------------------------------------------*/

/*
 * Context switch function used by the tick.  This must be identical to 
 * vPortYield() from the call to vTaskSwitchContext() onwards.  The only
 * difference from vPortYield() is the tick count is incremented as the
 * call comes from the tick ISR.
 */
void vPortYieldFromTick( void ) __attribute__ ( ( naked ) );
void vPortYieldFromTick( void )
{
	portSAVE_CONTEXT();
	if( xTaskIncrementTick() != pdFALSE )
	{
		vTaskSwitchContext();
	}
	portRESTORE_CONTEXT();

	asm volatile ( "ret" );
}
/*-----------------------------------------------------------*/

/*
 * Setup timer 1 compare match A to generate a tick interrupt.
 */
static void prvSetupTimerInterrupt( void )
{
unsigned long ulCompareMatch;
unsigned char ucHighByte, ucLowByte;

	/* Using 16bit timer 1 to generate the tick.  Correct fuses must be
	selected for the configCPU_CLOCK_HZ clock. */

	ulCompareMatch = configCPU_CLOCK_HZ / configTICK_RATE_HZ;

	/* We only have 16 bits so have to scale to get our required tick rate. */
	ulCompareMatch /= portCLOCK_PRESCALER;

	/* Adjust for correct value. */
	ulCompareMatch -= ( unsigned long ) 1;

	/* Setup compare match value for compare match A.  Interrupts are disabled 
	before this is called so we need not worry here. */
	ucLowByte = ( unsigned char ) ( ulCompareMatch & ( unsigned long ) 0xff );
	ulCompareMatch >>= 8;
	ucHighByte = ( unsigned char ) ( ulCompareMatch & ( unsigned long ) 0xff );
	OCR1AH = ucHighByte;
	OCR1AL = ucLowByte;

	/* Setup clock source and compare match behaviour. */
	ucLowByte = portCLEAR_COUNTER_ON_MATCH | portPRESCALE_64;
	TCCR1B = ucLowByte;

	/* Enable the interrupt - this is okay as interrupt are currently globally
	disabled. */
	ucLowByte = TIMSK;
	ucLowByte |= portCOMPARE_MATCH_A_INTERRUPT_ENABLE;
	TIMSK = ucLowByte;
}
/*-----------------------------------------------------------*/


#if configUSE_PREEMPTION == 1

	/*
	 * Tick ISR for preemptive scheduler.  We can use a naked attribute as
	 * the context is saved at the start of vPortYieldFromTick().  The tick
	 * count is incremented after the context is saved.
	 */
	void TIMER1_COMPA_vect( void ) __attribute__ ( ( signal, naked ) );
	void TIMER1_COMPA_vect( void )
	{
		vPortYieldFromTick();
		asm volatile ( "reti" );
	}
#else

	/*
	 * Tick ISR for the cooperative scheduler.  All this does is increment the
	 * tick count.  We don't need to switch context, this can only be done by
	 * manual calls to taskYIELD();
	 */
	void TIMER1_COMPA_vect( void ) __attribute__ ( ( signal ) );
	void TIMER1_COMPA_vect( void )
	{
		xTaskIncrementTick();
	}
#endif
	


/*
 * Tickless sleeping mode.
 * (C) Andrzej Michalec 2013, http://amichalec.net
 *
 * Implementation focused on maximum power saving and average accuracy. It requires
 * hardware having watch 32.768Hz crystal connected to TOSC0-TOSC1 and uses Timer0
 * for asynchronous clock.
 *
 * For long periods it uses deep "power save" sleep mode with only async 32.678
 * clock running and microampere current consumption. For requested sleep time
 * close to wakeup lag of main oscillator (ranging from micro to milliseconds)
 * it uses "extended standby" with main clock kept running at price of
 * sub-miliampere current consumption.
 *
 * To increase sleep time precision Timer0 clock prescaler is computed dynamically
 * Such extra calculation steps introduce overhead visible when sleep is requested
 * very frequently for short periods (like 5 ticks at 1000Hz). It is advised to add
 * configEXPECTED_IDLE_TIME_BEFORE_SLEEP parameter in FreeRTOSConfig.h if scheduler
 * time slicing is very granular.
 *
 * Tickless idle with 32 bit ticks requires extra ~1100 bytes of flash and for 16 bit
 * ticks around 900 bytes.
 */
#if configUSE_TICKLESS_IDLE == 1
	#define TIMER0_FOSC			32768

	static portTickType		deepSleepMinTicks;
	static volatile bool	countingCompleted;

	typedef struct {
		uint8_t			prescalerMask;
		uint32_t		count;
	} presc_t;

	static portTickType prvCalculateDeepSleepMinTicks();
	static void prvSleep();
	static void prvSleepExactTime( portTickType xExpectedIdleTime );
	static void prvSetupTimerAsync( void );
	static presc_t prvCalculatePrescaler( portTickType xExpectedIdleTime );

	#if configUSE_TICKLESS_IDLE_DEBUG == 1
		#define UDBG(variable) printf_P(PSTR(#variable " = %u\r"), variable);
		#define UDBG_NEXT() printf_P(PSTR("--------\r"));
	#else
		//empty implementations
		#define UDBG(variable)
		#define UDBG_NEXT()
	#endif


	/*
	 * Setup Timer0 async mode with low-freq crystal.
	 */
	static void prvSetupTimerAsync( void )
	{
		deepSleepMinTicks = prvCalculateDeepSleepMinTicks();
		UDBG(deepSleepMinTicks);
		TCCR0 = 0;
		TCCR0 |= _BV(WGM01);	/* CTC compare mode, prescaler stopped */
		TIMSK &= ~_BV(OCIE0);	/* disable compare interrupt */
		ASSR |= _BV(AS0);		/* set async mode */
		/* All subsequent operations on Timer0 register are dynamic,
		 * see prvSleepExactTime() */
	}

	__attribute__((weak)) void vPortSuppressTicksAndSleep( portTickType xExpectedIdleTime )
	{
		UDBG_NEXT();
		portENTER_CRITICAL();
		eSleepModeStatus status = eTaskConfirmSleepModeStatus();

		if (status == eNoTasksWaitingTimeout) {
			/* It is not necessary to configure an interrupt to bring the
			   microcontroller out of its low power state at a fixed time in the
			   future. */
			set_sleep_mode(SLEEP_MODE_PWR_SAVE);
			/* We can sleep in critical section because on exit real
			   I-bit of SREG is restored and prvSleep() do SEI on its own. */
			prvSleep();
			portEXIT_CRITICAL();

		} else if (status == eAbortSleep) {
			/* If a context switch is pending or a task is waiting for
			   the scheduler to be unsuspended then abandon the low
			   power entry. */
			portEXIT_CRITICAL();

		} else {
			portEXIT_CRITICAL();
			prvSleepExactTime(xExpectedIdleTime);
		}
	}


	/*
	 * Calculate minimum sleep time rationalizing deeper sleep
	 * that requires more time to wake up.
	 */
	static portTickType prvCalculateDeepSleepMinTicks()
	{
		/* Reading sys clock ticks to wake-up depending on oscillator used.
		   See atmega128a datasheet table 8-4 */
		uint8_t lfuse = boot_lock_fuse_bits_get( GET_LOW_FUSE_BITS );
		uint16_t wakeCK;
		if ( ~(lfuse & (1<<0)) & ( ~(lfuse & (1<<4)) | ~(lfuse & (1<<5)) ) ) {
			/* CKSEL0 (bit 0) is programmed and any of SUT0 (bit 4) or SUT1 (bit 5)
			   is programmed then 16K cycles. */
			wakeCK = 16000;
		} else if ( (lfuse & (1<<0)) & (lfuse & (1<<5)) ) {
			/* CKSEL0 is not-programmed and SUT1 is not-programmed */
			wakeCK = 258;
		} else {
			wakeCK = 1000;
		}
		UDBG(wakeCK);
		/* Expected sleep time (in ticks) has to be N-times longer than wake time.
		   Factor of 10 is to lower error of sleep count.
		   Example: for 12MHz clock crystal needs 16k cycles = 1.33ms to work again
		   so at 1000Hz ticks freq, we need at least 13.3ms so 14 ticks. */
		return ((uint32_t)configTICK_RATE_HZ * 10) / (configCPU_CLOCK_HZ / wakeCK);
	}

	/* Find the best Timer0 prescaler to sleep as long as possible with 8-bit
	 * precision. This happen for maximum possible prescaled frequency with
	 * overflow later than  requested sleep time i.e. when
	 * 		max(fpre) < tick_rate*256/expected
	 * Then counter should not exceed
	 * 		countMax = floor(expected*fpre/ticks_rate)
	 * Storing data in struct for further calculations.
	 */
	inline static presc_t prvCalculatePrescaler( portTickType xExpectedIdleTime )
	{
		presc_t pr;
		uint32_t fexp = ( (uint32_t)configTICK_RATE_HZ * 0xFF ) / xExpectedIdleTime;
		uint16_t fpre[] = {TIMER0_FOSC/1, TIMER0_FOSC/8, TIMER0_FOSC/32,
				TIMER0_FOSC/64, TIMER0_FOSC/128, TIMER0_FOSC/256, TIMER0_FOSC/1024};
		uint8_t size = sizeof(fpre)/sizeof(fpre[0]);
		pr.prescalerMask=0;
		for (uint8_t i=0; i < size; i++) {
			pr.prescalerMask++;
			if (fpre[i] < fexp) break;
		}
		/* Count maximum */
		pr.count = ((uint32_t)xExpectedIdleTime * fpre[pr.prescalerMask-1])
				/ configTICK_RATE_HZ;
		UDBG(pr.prescalerMask);
		UDBG(pr.count);
		return pr;
	}

	static void prvSleep()
	{
		/* Re-enabling interrupts to awake and go to sleep*/
		sleep_enable();
		sei();
		sleep_cpu();
		/* Sleeps here until awaken, then continues */
		sleep_disable();
		cli();
	}

	inline static void prvSleepExactTime( portTickType xExpectedIdleTime )
	{
		UDBG(xExpectedIdleTime);
		presc_t pr = prvCalculatePrescaler(xExpectedIdleTime);
		TCNT0 = 0;
		OCR0 = (pr.count > 0xFF) ? 0xFF : pr.count;
		TCCR0 &= ~0x07;			/* initially timer stopped (cut off prescaler) */

		/* configPRE_SLEEP_PROCESSING() can set its parameter to 0
		   to indicate that its implementation contains its own wait
		   for interrupt or wait for event instruction.  However,
		   the original expected idle time variable must remain
		   unmodified, so a copy is taken. */
		portTickType expectedIdleTimeCopy = xExpectedIdleTime;
		configPRE_SLEEP_PROCESSING(expectedIdleTimeCopy);
		portTickType preSleepTickCount = xTaskGetTickCount();
		if (expectedIdleTimeCopy > 0) {
			if (expectedIdleTimeCopy > deepSleepMinTicks) {
				set_sleep_mode(SLEEP_MODE_PWR_SAVE);
			} else {
				set_sleep_mode(SLEEP_MODE_EXT_STANDBY);
			}
			/* Enable TIMER0_COMP_vect interrupt with right prescaler */
			TIMSK |= _BV(OCIE0);
			TCCR0 |= (0x07 & pr.prescalerMask);
			/* Ticks should be stopped, however in sleep mode Timer1 is
			   not clocked so no extra step is required. */
			/* Wait for update of all modifier asynchronous registers,
			   otherwise no interrupt will occur */
			while(ASSR & (_BV(TCN0UB) | _BV(TCR0UB) | _BV(OCR0UB)));
			/* We can sleep in critical section because on exit real
			   I-bit of SREG is restored and prvSleep() do SEI on its own. */
			portENTER_CRITICAL();
			countingCompleted = false;
			prvSleep();
			/* Disable TIMER0_COMP_vect interrupt and stop timer
			   (cut off prescaler) */
			TIMSK &= ~_BV(OCIE0);
			TCCR0 &= ~0x07;
			portEXIT_CRITICAL();
		}
		/* Post sleep hook. */
		configPOST_SLEEP_PROCESSING(xExpectedIdleTime);

		portTickType completedTicks;
		/* Before checking TCNT0 after wakeup we need to wait until it
		   is ready. TCCR0 was touched so we wait for completion. */
		while(ASSR & _BV(TCR0UB));

		if (countingCompleted) {
			if (pr.count <= 0xFF) {
				/* MCU slept all expected time and it fit into 8-bit counter. */
				completedTicks = xExpectedIdleTime;
			} else {
				/* Initial calculation wanted MCU sleep longer than 8-bit counter can,
				   so we have to calculate fraction slept over. */
				completedTicks = ( ((uint32_t)xExpectedIdleTime * 0xFF) / pr.count ) +1;
			}
		} else {
			/* MCU was awaken for some other reason than TIMER0_COMP_vect
			   ISR so it did not count to OCR0. As ISR has not cleared TCNT0
			   we can read how far it got. +1 for rounding up */
			completedTicks = ( ((uint32_t)xExpectedIdleTime * TCNT0) / OCR0 ) +1;
		}
		/* Post-processing and ISR that has awaken MCU could take
		   some time. Delta of ticks latched before sleep and after
		   this point is added to completed. */
		completedTicks += ( xTaskGetTickCount() - preSleepTickCount );
		/* completeTicks cannot exceed given sleep time otherwise
		   jump in ticks will cause scheduler to miss waiting tasks. */
		if (completedTicks > xExpectedIdleTime) {
			completedTicks = xExpectedIdleTime;
		}
		UDBG(TCNT0);
		UDBG(completedTicks);
		vTaskStepTick(completedTicks);
	}

	ISR( TIMER0_COMP_vect )
	{
		countingCompleted = true;
	}

#endif /* configUSE_TICKLESS_IDLE == 1 */

